##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Run ALAMO.
"""
import sys
import collections
import numpy as np
import os
#
import alamopy
from alamopy import almerror, almpywriter, almwriter
from alamopy.writethis import writethis
from alamopy.multos import deletefile, movefile


def doalamo(xdata, zdata,**kwargs):
    """ [almmodel] = doalamo(xdata,zdata, xvaldata, zvaldata,addopt=vals)

    Args:
        xdata (numpy.array or list[real])
        zdata (numpy.array or list[real)
        kwargs: Additional options may be specified and will be applied
                to the .alm
          -  example -  monomialpower=(1,2,3,4) 
          -  xlabels       : labels given to input variables
          -  zlabels       : labels given to outputs
          -  xval          : validaiton data for alamo
          -  zval          : response validation data for alamo
          -  modeler       : modeler value used in alamo
          -  solvemip      : force alamo to solve mip if gams is availible
          -  linfcns       : 0-1 option to include linear transformations
          -  expfcns       : 0-1 option to include exponential transformations
          -  logfcns       : 0-1 option to include logarithmic transformations
          -  sinfcns       : 0-1 option to include sine transformations
          -  cosfcns       : 0-1 option to include cosine transformations
          -  monomialpower : list of monomial powers
          -  multi2power   : list of binomial powers
          -  multi3power   : list of trinomials
          -  ratiopower    : list of ratio powers
          -  almname       : specify a name for the .alm file
          -  savescratch   : saves .alm and .lst
          -  savetrace     : saves trace file
          -  expandoutput  : add a key to the output dictionary for the output
                             (must be on for inputs>1)
          -  almopt        : direct text appending
                             the option almopt=<file> will append a file to the
                             end of the .alm and can be used to facilitate
                             direct access to the .alm (no current checks)
    Returns:
        dict: An ALAMO model with the following keys
          -  'model'    : algebraic form of model
          -  'f(model)' : a callable lambda function
          -   Syntac is depended on expandout
               syntax => almmodel['f(model)']['out'](inputs,sep,by,comma)
                         almmodel['f(model)'](inputs,sep,by,comma)
          -  'ssr'      : SSE on training set provided
          -  'R2'       : R2 on training set provided
          -  'ssrval'   : SSE on testing set if provided
          -  'R2val'    : R2 on testing set if provided
    """
    data, debug = alamopy.data, alamopy.debug

    # patched together validation data check
    if 'xval' in kwargs.keys():
        vargs = (kwargs['xval'], kwargs['zval'])
    else:
        vargs = ()

    checkinput(data, xdata, zdata, vargs, kwargs, debug)

    # Begin to pull information from the training data
    data['almopts']['ndata']=np.shape(xdata)[0]
    data['almopts']['ninputs']=np.shape(xdata)[1]
    
    # Check training data
    if len(np.shape(zdata)) == 1:
        zdata = np.reshape(zdata, (data['almopts']['ndata'],1))
        data['almopts']['noutputs'] = 1
    else:
        data['almopts']['noutputs'] = np.shape(zdata)[1]
    if (np.shape(zdata)[0] != data['almopts']['ndata']):
        almerror('p1')
    elif (np.shape(xdata)[0] != data['almopts']['ndata']):
        almerror('p1')
    
    # Check for validation set
    if vargs != ():
        debug['validation']=True
        xvaldata=vargs[0]
        zvaldata=vargs[1]
        temp = np.shape(xvaldata)
        data['almopts']['nvaldata']=temp[0]
        if (len(np.shape(zvaldata)) == 1):
            zvaldata = np.reshape(zvaldata, (data['almopts']['nvaldata'],1))
        if (temp[1] != data['almopts']['ninputs']):
            writethis('Number of input variables inconsistent between x and xval')
            almerror('p2')
        temp = np.shape(zvaldata)
        if (temp[0] != data['almopts']['nvaldata'] or temp[1] != data['almopts']['noutputs']):
            writethis('Problem with zval')
            almerror('p2')

    # Get labels strucutres input data regardless of input type
    getlabels(data, debug, kwargs)

    # Parse additional input options
    # The 'pargs' library is used to keep track of options a user has availible
    for arg in kwargs.keys():
        if arg in data['pargs']['almopts']:
            data['almopts'][arg]=kwargs[arg]
        elif arg in data['pargs']['almlstopts']:
            data['almlstopts'][arg]=list()
            for term in list([kwargs[arg]]):
                data['almlstopts'][arg].append(term)
        elif arg in data['pargs']['almstropts']:
            data['almstropts'][arg]=kwargs[arg]
        elif arg in data['pargs']['almset4']:
            data['almset4'][arg]=kwargs[arg]
        elif arg in debug['pargs']:
            debug[arg] = kwargs[arg]
        else:
            if (arg not in (['xlabels', 'zlabels'])):
                sys.stdout.write('Problem with option : '+arg)
                almerror('p3')

    # Specific check to see if the labels of the response variables should be used in the output dictionary
    # This is important for systematic testing vs. single model input
    if debug['outkeys'] == False:
        # outkeys are specified to be used
        if data['almopts']['noutputs'] > 1:
            #'Must use outkeys for multiple outputs'
            writethis('outkeys set to TRUE for multiple outputs')
            debug['outkeys']=True

    # Construct xmin and xmax vector based on training data if not provided
    if ('xmin' not in kwargs.keys()):
        writethis('Xmin and Xmax are not provided, they will be calculated from the training data')
        xmin=''
        xmax=''
        for i in range(data['almopts']['ninputs']):
            tn = debug['bignum']
            tx=-1*debug['bignum']
            for j in range(data['almopts']['ndata']):
                if (float(xdata[j][i]) < float(tn)):
                    tn = xdata[j][i]
                if (float(xdata[j][i]) > float(tx)):
                    tx = xdata[j][i]
            xmin=xmin+str(tn)+' '
            xmax=xmax+str(tx)+' '
        data['almset4']['xmax']=xmax
        data['almset4']['xmin']=xmin

    if debug['validation']:
        alamopy.almwriter(data,debug,(xdata,zdata,xvaldata,zvaldata),kwargs)
    else:
        alamopy.almwriter(data,debug,(xdata,zdata),kwargs)


    # Call alamo from the terminal
    if debug['showalm']:
        os.system(debug['almloc']+" "+str(data['almstropts']['almname']))
    else:
        writethis('Calling ALAMO now:\n')
        os.system(debug['almloc']+" "+str(data['almstropts']['almname'])+" > logscratch")


    data['results']={}

    #Check to see if additional data was sampeled and add it
    if 'sampler' in kwargs.keys():
        with open(data['almstropts']['almname'].split('.')[0]+".lst") as infile, open('awkres','w') as outfile:
            copy = False
            for line in infile:
                if "Errors on observed data points" in line.strip():
                    copy = True
                elif "Maximum absolute errors" in line.strip():
                    copy = False
                elif copy:
                    outfile.write(line)
        f = open('awkres')
        lf=f.read()
        f.close()
        lf2=lf.split('\n')
        lf2 = lf2[1:-1]
        sys.stdout.write('Updating number of training points from '+str(data['almopts']['ndata'])+' to '+str(len(lf2))+'\n')
        data['almopts']['ndata']=len(lf2)
        xdata=np.zeros([data['almopts']['ndata'],data['almopts']['ninputs']])
        zdata=np.zeros([data['almopts']['ndata'],data['almopts']['noutputs']])
        for i in range(len(lf2)):
            lf3=lf2[i].split(' ')
            while '' in lf3: lf3.remove('')
            for j in range(data['almopts']['ninputs']):
                xdata[i][j]=float(lf3[j])
            for j in range(data['almopts']['noutputs']):
                zdata[i][j]=float(lf3[data['almopts']['ninputs']+j])
        deletefile("awkres")
        xdata=xdata
        zdata=zdata
    if debug['expandoutput']:
        data['results']['xdata']=xdata
        data['results']['zdata']=zdata
        data['results']['xlabels']=data['almlabs']['savexlabels']
        data['results']['zlabels']=data['almlabs']['savezlabels']
    try:  
        import sympy
        from sympy.parsing.sympy_parser import parse_expr
        from sympy import symbols, lambdify
    except:
        writethis('Cannot install sympy')

    if (debug['expandoutput'] == False):
        data['results']['model']={}
        data['results']['f(model)']={}
    else:
        for key in list(['model','f(model)','ssr','R2','size','rmse','nbas','totaltime','olrtime','miptime','clrtime','othertime','version','status','madp','numolr','nummip','numclr','ninputs']):
            data['results'][key]=collections.OrderedDict()
        if (len(vargs)>0):
            data['results']['ssrval']=collections.OrderedDict()
            data['results']['R2val']=collections.OrderedDict()
            data['results']['rmseval']=collections.OrderedDict()
            data['results']['madpval']=collections.OrderedDict()

    # Open the trace file and pull appropriate results
    trace_file = data['almstropts']['tracefname']
    try:
        lf = open(trace_file).read()
    except IOError as err:
        raise almerror.AlamoError('Cannot read from trace file "{}": {}'
                                  .format(trace_file, err))
    lf2 = lf.split('\n')
    tkeys=lf2[0].split(',')
    kl1=list(['ssr','rmse','R2','size','nbas','totaltime','olrtime','miptime','clrtime','othertime','version','status','madp','numolr','nummip','numclr','ninputs'])
    kl2=list([' SSE',' RMSE',' R2',' ModelSize',' nBasInitAct',' TotalTime', ' OLRTime',' MIPTime',' CLRTime',' OtherTime', ' AlamoVersion',' AlamoStatus',' MADp',' numOLRs',' NumMIPs',' numCLRs',' NINPUTS'])
    # Construct results for training data (&val if provided)
    ln=1
    wlparam=data['almopts']['noutputs']
    if (len(vargs) > 0):
        wlparam=2*wlparam
    else:
        wlparam=wlparam+1
    while ln < wlparam:
        lf3=lf2[ln].split(',')
        # Reapply the saved labels for the output
        model=lf3[tkeys.index(' Model')]
#    for label in data['almlabs']['savexlabels']:
        for i in range(data['almopts']['ninputs']):
            label=data['almlabs']['xlinks'][i][0]
            # Now is a convenient time to collect information that will be used in the
            # confidence interval analysis
            model=model.replace(str(label),str(data['almlabs']['xlinks'][i][1]))
        for i in range(data['almopts']['noutputs']):
            label=data['almlabs']['zlinks'][i][0]
            model=model.replace(str(label),str(data['almlabs']['zlinks'][i][1]))
        # determine which output label to write 
        # if debug['outkeys'] == True use olab as a key if not dont
        if debug['outkeys']:
            olab = model.split('=')[0]
            olab=olab.replace(' ','')
            data['results']['model'][olab]=model
            #Record tokenized model for each output
            data['results']['f(model)'][olab] = lambdify([symbols(data['almlabs']['savexlabels'])], parse_expr(model.split('=')[1].replace('^','**')), "numpy")
        else:
            data['results']['model']=model
            data['results']['f(model)']=lambdify([symbols(data['almlabs']['savexlabels'])], parse_expr(model.split('=')[1].replace('^','**')), "numpy")
        if debug['expandoutput']:
            if debug['outkeys']:
                for i in range(len(kl1)):
                    data['results'][kl1[i]][olab]=lf3[tkeys.index(kl2[i])]
                # Check for validation set
                if len(vargs)>0:
                    lf3=lf2[2].split(',')
                    data['results']['ssrval'][olab]=lf3[tkeys.index(' SSE')]
                    data['results']['R2val'][olab]=lf3[tkeys.index(' R2')]
                    data['results']['rmseval'][olab]=lf3[tkeys.index(' RMSE')]
                    data['results']['madpval'][olab]=lf3[tkeys.index(' MADp')]
            else:
                for i in range(len(kl1)):
                    data['results'][kl1[i]]=lf3[tkeys.index(kl2[i])]
                # Check for validation set
                if len(vargs)>0:
                    lf3=lf2[2].split(',')
                    data['results']['ssrval']=lf3[tkeys.index(' SSE')]
                    data['results']['R2val']=lf3[tkeys.index(' R2')]
                    data['results']['rmseval']=lf3[tkeys.index(' RMSE')]
                    data['results']['madpval']=lf3[tkeys.index(' MADp')]
        else:
            if debug['outkeys']:
                data['results']['ssr'][olab]=lf3[tkeys.index(kl2[0])]
            else:
                data['results']['ssr']=lf3[tkeys.index(kl2[0])]
        ln=ln+1

    alamopy.almpywriter(data, debug['cvfun'])
    #add <>alm.py to results dict
    data['results']['pymodel'] = data['almstropts']['almname'].split('.')[0]+'alm' 
    # Delete files
    if not debug['savescratch']:
        deletefile(str(data['almstropts']['almname'])+" "+str(data['almstropts']['almname']).split('.')[0]+".lst")
    if not debug['savetrace']:
        deletefile(data['almstropts']['tracefname'])

    return data['results']


def checkinput(data, xdata, zdata, vargs, kwargs, debug):
    """Check the input data input doalamo for errors.
    """
    t = np.zeros([1, 1])
    kk = kwargs.keys()
    # First check vargs for validation set viability
    if len(vargs) > 0 and len(vargs) != 2:
        raise almerror.AlamoInputError('Validation Xdata and Zdata must be'
                                       'specified in the following form: '
                                       'alamopy.doalamo(x,z,xval,zval)')
    elif len(vargs) == 2:
        if type(vargs[0]) != type(t) or type(vargs[1]) != type(t):
            raise almerror.AlamoInputError('Validation data must be provided '
                                           'as numpy arrays')
    if type(xdata) != type(t) or type(zdata) != type(t):
        raise almerror.AlamoInputError('Xdata and Zdata must be provided'
                                       'as numpy arrays')
    for arg in kk:
        if arg not in data['pargs']['almopts'] + data['pargs']['almstropts'] +\
                data['pargs']['almlstopts'] + data['pargs']['almset4'] + \
                debug['pargs'] + ['xlabels', 'zlabels']:
            raise almerror.AlamoInputError('The following argument given to'
                                           'doalamo() is not understood: {}'
                                           .format(arg))
    if 'expandoutput' in kk:
        debug['expandoutput'] = kwargs['expandoutput']
    if 'almpath' in kk:
        debug['almloc'] = kwargs['almpath']
    if 'gamspath' in kk:
        debug['gamsloc'] = kwargs['gamspath']
    if 'cvfun' in kk:
        debug['cvfun'] = kwargs['cvfun']

def getlabels(data, debug, kwargs):
    # This function generates labels if they are not provided

    # Check to see if labels have been specified before we generate them
    if ( 'xlabels' in kwargs.keys()):
        data['almlabs']['savexlabels']=kwargs['xlabels']
    else:
        # Make savexlabels
        temp=list()
        for i in range(data['almopts']['ninputs']):
            temp.append("x"+str(i+1))
        data['almlabs']['savexlabels']=temp
    if 'zlabels' in kwargs.keys():
        data['almlabs']['savezlabels']=kwargs['zlabels']
    else:
        # Make savezlabels
        temp=list()
        for i in range(data['almopts']['noutputs']):
            temp.append("z"+str(i+1))
        data['almlabs']['savezlabels']=temp

    # create temp labels for alm file, avoid checking user labels for alm rules
    data['almlabs']['xlinks']=[0 for i in range(data['almopts']['ninputs'])]
    data['almlabs']['zlinks']=[0 for i in range(data['almopts']['noutputs'])]    
    makelabs(data,debug,'ninputs')
    makelabs(data,debug,'noutputs')

def makelabs(data, debug, param):
    """
    -  Constructs labels for alamo
    param = 'ninputs' or 'noutputs'
    """
    temp=list([])
    for i in range(data['almopts'][param]):
        if param == 'ninputs':
            data['almlabs']['xlinks'][i]=['a','b']
            temp.append("almx"+str(i+1)+'d')
            data['almlabs']['xlinks'][i][0]='almx'+str(i+1)+'d'
            data['almlabs']['xlinks'][i][1]=data['almlabs']['savexlabels'][i]
            key='xlabels'
        elif param == 'noutputs':
            data['almlabs']['zlinks'][i]=['a','b']
            temp.append("almz"+str(i+1)+'d')
            data['almlabs']['zlinks'][i][0]='almz'+str(i+1)+'d'
            data['almlabs']['zlinks'][i][1]=data['almlabs']['savezlabels'][i]
            key='zlabels'
    data['almlstopts'][key]=temp
